// PieGraph.cpp : implementation file
//
// CrackUp version 1.0
// Copyright  1998 Ziff-Davis, Inc.
// First published in PC Magazine, US Edition, November 3, 1998
// Written by Bob Flanders and Gregory A. Wolking

#include "stdafx.h"
#include "PieGraph.h"
#include <math.h>

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif


/////////////////////////////////////////////////////////////////////////////
// CPieGraph

CPieGraph::CPieGraph()
{									// Set defaults:
	m_fPercent = 0.0f;				// zero percent.
	m_crPrimary = 0x00ff0000;		// Primary slice color = blue
	m_crSecondary = 0x00ffff20;		// Secondary slice color = cyan
	m_crLines = 0x00000000;			// Line color = black
	m_crBackground = 0x00ffffff;	// Background color = white
	m_hCurrentBitmap = NULL;
}

CPieGraph::~CPieGraph()
{
	if (m_hCurrentBitmap)					// If our bitmap is still allocated,
		DeleteObject(m_hCurrentBitmap);		// delete it.
}


BEGIN_MESSAGE_MAP(CPieGraph, CStatic)
	//{{AFX_MSG_MAP(CPieGraph)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CPieGraph message handlers

////////////////////////////////////////////////////////
// CPieGraph public member functions.


// Member function to set new percentage and update the graph.
void CPieGraph::SetPercentage(float fPercentage)
{
	// Do nothing if percentage is out of range.
	if (fPercentage > 1.0f || fPercentage < 0.0f)
		return;
	m_fPercent = fPercentage;		// Set member variable.
	DrawBitmap();					// Draw and install new bitmap.
}

// Overload allows using setting percentage using an int from 0 to 100
void CPieGraph::SetPercentage(int iPercentage)
{
	if (iPercentage > 100 || iPercentage < 0)
		return;
	SetPercentage(iPercentage / 100.0f);	// Call float version.
}

//////////////////////////////////////////////////////////
// CPieGraph helpers

void CPieGraph::DrawSlice(CDC *pDC,					// context to draw in
		 					int		iX,				// center point x
							int		iY,				// center point y
							DWORD	dwRadius,		// radius of line
							float	fStartDegrees,	// start degrees for arc
							float	fSweepDegrees,	// end degrees for arc
							int		iHeight,		// height for 3d
							BOOL	bDrawLines,		// TRUE to draw lines
							CPoint	*lpptP,			// pointer to end CPoint 
							CPoint  *lpptS)
{
	int		iXStart,										// end point of starting radial line
			iYStart,										// .. x,y coordinates
			iXEnd,											// end point of ending radial line
			iYEnd;											// .. x,y coordinates

	float	fStartRadians,									// Start angle in radians
			fEndRadians,									// End angle in radians
			fX, fY;											// work floats

	#define PIE_X_2			6.283184f						// pi*2 as float

	// Calculate start/end angles 
	// .. in radians
	if (fSweepDegrees > 0.0f)								// q. positive sweep degrees?
	{														// a. yes .. 
		fStartRadians = ((fStartDegrees/360.0f) * PIE_X_2);	// .. calc start radians
		fEndRadians = (((fStartDegrees+fSweepDegrees)		// .. and ..	
								/ 360.0f) * PIE_X_2);		// .. end radians	
	} 
	else													// otherwises
	{
		fStartRadians = (((fStartDegrees+fSweepDegrees)		// .. calculate 
								/ 360.0f) * PIE_X_2);		// .. for other direction
		fEndRadians =  ((fStartDegrees/360.0f) * PIE_X_2);	// .. start and end
	}
	fX = (float) dwRadius * (float) cos(fStartRadians);		// calc x coordinate offset
	fY = (((float) dwRadius)/2.0f)							// .. and y offset
		   * (float) sin(fStartRadians);
	iXStart = iX + (int)(fX);								// calculate start point x coordinate
	iYStart = iY - (int)(fY);								// .. and the y coordinate
	if (lpptS != NULL)										// q. point requested?
	{														// a. yes ..
		lpptS->x = iXStart;									// .. set the start point x
		lpptS->y = iYStart;									// .. and the y coordinates
	}
	fX = (float) dwRadius * (float) cos(fEndRadians);		// calc x coordinate offset
	fY = (((float) dwRadius)/2.0f)							// .. and y offset
		   * (float) sin(fEndRadians);
	iXEnd = iX + (int)(fX);									// calculate end point x coordinate
	iYEnd = iY - (int)(fY);									// .. and the y coordinate
	if (lpptP != NULL)										// q. point requested?
	{														// a. yes ..
		lpptP->x = iXEnd;									// .. set the end point x
		lpptP->y = iYEnd;									// .. and the y coordinates
	}
	if (!bDrawLines)										// q. lines wanted?
	{														// a. no ..
		MoveToEx(pDC->m_hDC, iXStart, iYStart, NULL);		// .. move to the arc start point
	}
	if (iHeight && iYStart >= iY)							// q. any height & below center Y?
	{														// a. yes ..
		pDC->LineTo(iXStart, iYStart + iHeight);				// .. draw side line
		MoveToEx(pDC->m_hDC, iXStart, iYStart, NULL);		// .. return to start point
	}
	if (bDrawLines)											// Lines requested?
		pDC->Pie(iX - dwRadius, iY - (dwRadius/2),			// No -- draw entire pie slice.
				iX + dwRadius, iY + (dwRadius/2),
				iXStart, iYStart,
				iXEnd, iYEnd);
	else													// Otherwise, just draw an arc.
		pDC->Arc(iX - dwRadius, iY - (dwRadius/2),			// .. build a box ..
				iX + dwRadius, iY + (dwRadius/2),			// .. bounding the ellipse	
				iXStart, iYStart,							// .. and give the start ..
				iXEnd,   iYEnd);							// .. and end points of the arc
	MoveToEx(pDC->m_hDC, iXEnd, iYEnd, NULL);				// go to the end point
	if (iYEnd < iY)											// q. current y above center?
		iHeight = 0;										// a. no .. height lines shouldn't show
	if (iHeight)											// q. any height?
	{														// a. yes ..
		pDC->LineTo(iXEnd, iYEnd + iHeight);				// .. draw side line
	}
}

void CPieGraph::DrawPie(CDC* pDC)
{
#define FFS FLOODFILLSURFACE					// shorthand
	CBrush		brPrimary(m_crPrimary),			// brush for primary slice
				brSecondary(m_crSecondary),		// brush for secondary slice
				brBackground(m_crBackground),	// brush for background.
				*pOldBrush = NULL;				// pointer to original brush.
	CPen		penLines(PS_SOLID, 2, m_crLines),	// pen for drawing lines
				*pOldPen = NULL;				// pointer to original pen.
	float		fCurDeg = 0.0f,					// current point on the chart
				fPrimary,						// degrees of primary slice
				fSecondary,						// degrees of secondary slice
				fX;								// work float
	int			iCtrX,							// center x coordinate
				iCtrY,							// center y coordinate
				iWidth,							// width of client rectangle
				iHeight,						// height of client rectangle
				iRadius;						// radius of the chart		
	RECT		rectX;							// work rectangle
	CPoint		ptFF;							// front fill point
	
	pDC->SetMapMode(MM_TEXT);
	pOldPen = pDC->GetCurrentPen();			// Save default pen
	pOldBrush = pDC->GetCurrentBrush();		// Save default brush.
	GetClientRect(&rectX);					// .. get screen based pie coordinates
	// **************************************************************************
	// Put a border around the pie chart area
	// **************************************************************************
	iWidth = rectX.right - rectX.left;						// calc width of rectangle
	iHeight = rectX.bottom - rectX.top;						// .. and height
	iCtrX = iWidth / 2;										// get half the width
	iCtrX += rectX.left;									// .. calc center x coord
	iCtrY = iHeight / 2;									// get half the height
	iCtrY += rectX.top;										// .. calc center y coord
	iRadius = (int) ((float)iWidth * .40f);					// use 40% of space for radius	
	fPrimary = 360 * (((int)(m_fPercent * 100.0f)) / 100.0f);	// Get percentage truncated to two decimal places.
	fSecondary = 360.0f - fPrimary;							// Remainder of circle is secondary slice.
	pDC->FillRect(&rectX, &brBackground);					// Paint rectangle with background color.
	pDC->SelectObject(&penLines);							// Select pen for drawing lines.
	DrawSlice(pDC,				// draw the bottom half circle
			  iCtrX, iCtrY+10,	// .. using our center y+10
			  iRadius,			// .. and radius
			  180.0f, 180.0f,	// .. from degree 180 for 180 degrees
			  -10,				// .. draw the desscenders
			  FALSE);			// .. and no lines
	if (fPrimary)				// q. enough to show?
	{							
		
		pDC->SelectObject(&brPrimary);	// Select Fragged brush.
		DrawSlice(pDC,			// draw the fragged part
				  iCtrX, iCtrY,	// .. using our center
				  iRadius,		// .. and radius
				  180.0f,		// .. from 180 
				  fPrimary,		// .. for fragged degrees
				  10,			// .. draw descenders
				  TRUE,			// .. and lines
				  &ptFF);		// .. save end point
		if (fPrimary > 90)									// q. more than 90 degrees?
		{
			pDC->ExtFloodFill(iCtrX-2,						// a. yes .. fill center of front
							iCtrY + (iRadius/2) + 3,	
							m_crBackground,
							FFS);
		}
		else												// else
		{
			if (fPrimary > 10)
				pDC->ExtFloodFill(ptFF.x-2, ptFF.y + 3, m_crBackground, FFS);// .. fill smaller face area
		}
	}
	if (fSecondary)				// q. enough to show?
	{				
		fX = 180.0f - fSecondary;	// calc start position
		if (fX < 0)				// q. lower than zero?
			fX = 360.0f + fX;	// a. yes .. calc new start point
		pDC->SelectObject(&brSecondary);	// Select UnFragged brush.
		DrawSlice(pDC,			// draw the unfragged part
				  iCtrX, iCtrY,	// .. using our center
				  iRadius,		// .. and radius
				  fX,			// .. from calculated start address
				  fSecondary,		// .. for unfragged degrees
				  10,			// .. draw descenders
				  TRUE,			// .. and lines
				  NULL,			// .. discard end point
				  &ptFF);		// .. save start point
		if (fSecondary > 180.0f)								// q. Did it come round front?
		{
			if (fSecondary > 270.0f)							// a. yes .. through middle?
			{
				pDC->ExtFloodFill(iCtrX-2,					// .. fill from 
								iCtrY + (iRadius/2) + 5,	// .. .. center
								m_crBackground, FFS);					// .. .. .. of front
			}
			else if (fSecondary > 185.0f)		// q. enough to fill?
			{
				pDC->ExtFloodFill(ptFF.x+2,	// a. yes ..
								ptFF.y+5,	// .. fill from
								m_crBackground, FFS);	// .. .. descender
			}
		}
	}
	if (pOldPen)						// If original pen was saved,
		pDC->SelectObject(pOldPen);		// restore it.
	if (pOldBrush)						// If original brush was saved,
		pDC->SelectObject(pOldBrush);	// restore it.
}

// Uses current percentage to draw and install our bitmap.
void CPieGraph::DrawBitmap()
{
	CDC dc;							// Device context for drawing.
	CDC* pDC;						// Pointer to client device context.
	HBITMAP hOldBitmap, hNewBitmap;	// Bitmap handles.
	RECT rect;						// work rectanle.

	GetClientRect(&rect);				// Get dimensions for bitmap.
	pDC = GetDC();						// Get pointer to client area device context.

	ASSERT(pDC != NULL);

	dc.CreateCompatibleDC(pDC);			// Create compatible context for in-memory drawing.
	// Create bitmap compatible with client context
	hNewBitmap = CreateCompatibleBitmap(pDC->m_hDC, rect.right, rect.bottom);

	ASSERT(hNewBitmap != NULL);

	ReleaseDC(pDC);									// Done with client context.
	dc.SelectObject(hNewBitmap);					// Select bitmap into compatible context.
	DrawPie(&dc);									// Draw the graph in the bitmap.
	hOldBitmap = CStatic::SetBitmap(hNewBitmap);	// Attach new bitmap, get handle to the old bitmap.
	if (hOldBitmap != NULL)							// If there was an old bitmap,
		if (hOldBitmap == m_hCurrentBitmap)			// and we created it,
			DeleteObject(hOldBitmap);				// delete it.
	m_hCurrentBitmap = hNewBitmap;					// Save handle to new bitmap.
}

// Since this control creates and manages its own bitmap,
// this override makes sure that our own bitmap is destroyed
// if user wants to attach his own "foreign" bitmap to the control
HBITMAP CPieGraph::SetBitmap(HBITMAP hBitmap)
{
	HBITMAP hOldBitmap;
	
	hOldBitmap = CStatic::SetBitmap(hBitmap);	// Set the new bitmap, get handle to the old bitmap.
	if (hOldBitmap != NULL)						// Did control have a bitmap?
	{
		if (hOldBitmap == m_hCurrentBitmap)		// If so, was it ours?
		{
			DeleteObject(m_hCurrentBitmap);			// If so, delete the object
			hOldBitmap = m_hCurrentBitmap = NULL;	// and null the handles.
		}
	}
	return hOldBitmap;	// Return old bitmap handle to caller.
}
